from django import template
from django.template import Node, Context
from django.template.base import TextNode, NodeList
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe


def cotton_component(parser, token):
    bits = token.split_contents()
    tag_name = bits[0]
    template_path = bits[1]

    kwargs = {}
    for bit in bits[2:]:
        key, value = bit.split("=")
        kwargs[key] = value

    nodelist = parser.parse(("end_cotton_component",))
    parser.delete_first_token()

    return CottonComponentNode(nodelist, template_path, kwargs)


class CottonComponentNode(Node):
    def __init__(self, nodelist, template_path, kwargs):
        self.nodelist = nodelist
        self.template_path = template_path
        self.kwargs = kwargs
        self.named_slots = {}

    def render(self, context):
        local_context = context.flatten()

        resolved_kwargs = {}
        for key, value in self.kwargs.items():
            try:
                resolved_value = template.Variable(value).resolve(context)
                resolved_kwargs.update({key: resolved_value})
            except TypeError:
                resolved_kwargs.update({key: value})
            except template.VariableDoesNotExist:
                resolved_kwargs.update({key: value})

        # Add the remainder as the default slot
        rendered = self.nodelist.render(context)
        local_context.update({"slot": rendered})

        slots = context.get("cotton_slots", {})
        local_context.update(slots)

        local_context.update(resolved_kwargs)

        # save attrs dict to context for
        local_context.update({"attrs_dict": resolved_kwargs})

        # Provide all of the attrs as a string to pass to the component
        # todo: should we be having defined variables in the component and this string only includes the remainder?
        # because of the issue where django changes <div {{attrs}}> to <div {{attrs}}="">, it leaves all attrs after the
        # initial, without quotes around the value
        def ensure_quoted(value):
            if value.startswith('"') and value.endswith('"'):
                return value
            else:
                return f'"{value}"'

        attrs = " ".join(
            [f"{key}={ensure_quoted(value)}" for key, value in self.kwargs.items()]
        )
        local_context.update({"attrs": mark_safe(attrs)})

        rendered = render_to_string(self.template_path, local_context)

        # check for with vars to add
        cotton_vars = local_context.get("cotton_vars", {})
        if len(cotton_vars):
            with_statements = " ".join(
                [
                    f"{key}={value}"
                    for key, value in context.items()
                    if key in self.cotton_vars
                ]
            )
            rendered = "{% with " + with_statements + " %}" + rendered + "{% endwith %}"
            del local_context["cotton_vars"]

        return rendered



from django import template
from django.utils.safestring import mark_safe


def cotton_slot(parser, token):
    try:
        # Splitting by None == splitting by spaces.
        tag_name, slot_name = token.split_contents()
    except ValueError:
        raise template.TemplateSyntaxError(
            "%r name missing from x-slot" % token.contents.split()[0]
        )

    nodelist = parser.parse(("end_cotton_slot",))
    parser.delete_first_token()
    return CottonSlotNode(slot_name, nodelist)


class CottonSlotNode(template.Node):
    def __init__(self, slot_name, nodelist):
        self.slot_name = slot_name
        self.nodelist = nodelist

    def render(self, context):
        # Here, we add the rendered content to the context
        # Todo, we need to ensure these vars are scoped to the parent component only
        if "cotton_slots" not in context:
            context.update({"cotton_slots": {}})

        output = self.nodelist.render(context)

        # delete cotton_slots key for slot_name
        if self.slot_name in context["cotton_slots"]:
            del context["cotton_slots"][self.slot_name]

        context["cotton_slots"][self.slot_name] = mark_safe(output)
        return ""



from django import template

register = template.Library()


def cotton_var(parser, token):
    # Split the token to get variable assignments
    parts = token.split_contents()
    cotton_vars = {}
    for part in parts[1:]:
        key, value = part.split('=')
        cotton_vars[key] = value

    return CottonVarNode(cotton_vars)


register.tag("cotton_var", cotton_var)


class CottonVarNode(template.Node):
    def __init__(self, cotton_vars):
        self.cotton_vars = cotton_vars

    def render(self, context):
        resolved_vars = {}

        # if the same var is already set in context, it's being passed explicitly to override the cotton_var
        # if not, then we resolve it from the context
        for key, value in self.cotton_vars.items():
            # if key in context:
            #     resolved_vars[key] = context[key]
            #     continue
            try:
                resolved_vars[key] = template.Variable(value).resolve(context)
            except (TypeError, template.VariableDoesNotExist):
                resolved_vars[key] = value

        cotton_vars = {'cotton_vars': resolved_vars}

        # Update the global context directly
        context.update(resolved_vars)
        context.update(cotton_vars)
        context['cotton_vars'].update(resolved_vars)

        return ''



from django.views.generic import TemplateView
from . import views
from django.urls import path

app_name = 'django_cotton'

class NamedSlotInLoop(TemplateView):
    template_name = 'named_slot_in_loop.cotton.html'

    def get_context_data(self, **kwargs):
        return {
            'items': [
                {'name': 'Item 1'},
                {'name': 'Item 2'},
                {'name': 'Item 3'},
            ]
        }


urlpatterns = [
    path('parent', TemplateView.as_view(template_name='parent_test.cotton.html')),
    path('child', TemplateView.as_view(template_name='child_test.cotton.html')),
    path('self-closing', TemplateView.as_view(template_name='self_closing_test.cotton.html')),
    path('include', TemplateView.as_view(template_name='cotton_include.cotton.html')),
    path('playground', TemplateView.as_view(template_name='playground.cotton.html')),
    path('tag', TemplateView.as_view(template_name='tag.cotton.html')),
    path('named-slot-in-loop', NamedSlotInLoop.as_view()),
    path('test/compiled-cotton', views.compiled_cotton_test_view, name='test_compiled_cotton'),
    path('test/cotton', views.cotton_test_view, name='test_cotton'),
    path('test/native-extends', views.native_extends_test_view, name='test_native_extends'),
    path('test/native-include', views.native_include_test_view, name='test_native_include'),
    path('attribute-merging', views.attribute_merging_test_view, name='attribute_merging'),
    path('attribute-passing', views.attribute_passing_test_view, name='attribute_passing'),
]



import hashlib
import os
import re
import warnings

from django.template.loaders.base import Loader as BaseLoader
from django.core.exceptions import SuspiciousFileOperation
from django.template import TemplateDoesNotExist
from django.utils._os import safe_join
from django.template import Template
from django.core.cache import cache
from django.template import Origin
from django.conf import settings

from bs4.formatter import HTMLFormatter
from bs4 import BeautifulSoup
from bs4 import MarkupResemblesLocatorWarning

warnings.filterwarnings("ignore", category=MarkupResemblesLocatorWarning)


class Loader(BaseLoader):
    is_usable = True

    def __init__(self, engine, dirs=None):
        super().__init__(engine)
        self.cache_handler = TemplateCacheHandler()
        self.dirs = dirs

    def get_template(self, template_name, skip=None):
        """
        Call self.get_template_sources() and return a Template object for
        the first template matching template_name. If skip is provided, ignore
        template origins in skip. This is used to avoid recursion during
        template extending.
        """

        tried = []

        for origin in self.get_template_sources(template_name):
            if skip is not None and origin in skip:
                tried.append((origin, "Skipped to avoid recursion"))
                continue

            try:
                contents = self.get_contents(origin)
            except TemplateDoesNotExist:
                tried.append((origin, "Source does not exist"))
                continue
            else:
                return Template(
                    contents,
                    origin,
                    origin.template_name,
                    self.engine,
                )

        raise TemplateDoesNotExist(template_name, tried=tried)

    def get_contents(self, origin):
        # check if file exists, whilst getting the mtime for cache key
        try:
            mtime = os.path.getmtime(origin.name)
        except FileNotFoundError:
            raise TemplateDoesNotExist(origin)

        # check and return cached template
        cache_key = self.cache_handler.get_cache_key(origin.template_name, mtime)
        cached_content = self.cache_handler.get_cached_template(cache_key)
        if cached_content is not None:
            return cached_content

        # If not cached, process the template
        compiled_template = self._compile_template(origin.name)

        # Cache the processed template
        self.cache_handler.cache_template(cache_key, compiled_template)

        return compiled_template

    def _compile_template(self, template_name):
        try:
            with open(template_name, "r") as f:
                content = f.read()

            compiled = self._compile_cotton_to_django(content)

            cleaned = self._fix_django_attribute_parsing(compiled)

            return cleaned

        except FileNotFoundError:
            raise TemplateDoesNotExist(template_name)

    def _fix_django_attribute_parsing(self, contents):
        """Django's template parser adds ="" to empty attribute-like parts in any html-like node, i.e. <div {{ something }}> gets
        compiled to <div {{ something }}=""> Then if 'something' is holding attributes sets, the last attribute value is
        not quoted. i.e. model=test not model="test"."""
        cleaned_content = re.sub('}}=""', "}}", contents)
        return cleaned_content

    def get_dirs(self):
        return self.dirs if self.dirs is not None else self.engine.dirs

    def get_template_sources(self, template_name):
        """Return an Origin object pointing to an absolute path in each directory
        in template_dirs. For security reasons, if a path doesn't lie inside
        one of the template_dirs it is excluded from the result set."""
        if not template_name.endswith(".cotton.html"):
            raise TemplateDoesNotExist(template_name)
        for template_dir in self.get_dirs():
            try:
                name = safe_join(template_dir, template_name)
            except SuspiciousFileOperation:
                # The joined path was located outside of this template_dir
                # (it might be inside another one, so this isn't fatal).
                continue

            yield Origin(
                name=name,
                template_name=template_name,
                loader=self,
            )

    def _remove_whitespace_in_tags_for_bs4_happiness(self, html_content):
        # Regular expression to match Django template tags and remove whitespace
        pattern = re.compile(r"\{\{\s*(.*?)\s*\}\}")
        # Replace matched template tags with their trimmed content
        return pattern.sub(r"{{\1}}", html_content)

    def _compile_cotton_to_django(self, html_content):
        """Convert cotton <c-* syntax to {%."""

        html_content = self._remove_whitespace_in_tags_for_bs4_happiness(html_content)

        soup = BeautifulSoup(html_content, "html.parser")

        self._transform_named_slots(soup)
        self._transform_components(soup)

        return str(soup)

    def _transform_named_slots(self, soup):
        """Replace <c-slot> tags with the {% cotton_slot %} template tag"""
        for c_slot in soup.find_all("c-slot"):
            slot_name = c_slot.get("name", "").strip()
            inner_html = "".join(str(content) for content in c_slot.contents)

            cotton_slot_tag = (
                f"{{% cotton_slot {slot_name} %}}{inner_html}{{% end_cotton_slot %}}"
            )
            c_slot.replace_with(BeautifulSoup(cotton_slot_tag, "html.parser"))

        return soup

    def _transform_components(self, soup):
        """Replace <c-[component path]> tags with the {% cotton_component %} template tag"""
        for tag in soup.find_all(re.compile("^c-"), recursive=True):
            component_name = tag.name[2:]  # Remove the 'c-' prefix

            # Convert dot notation to path structure and replace hyphens with underscores
            path = component_name.replace(".", "/").replace("-", "_")

            # Construct the opening tag
            opening_tag = "{{% cotton_component cotton/{}.cotton.html".format(path)
            for attr, value in tag.attrs.items():
                if attr == "class":
                    value = '"' + " ".join(value) + '"'
                opening_tag += " {}={}".format(attr, value)
            opening_tag += " %}"

            # Construct the closing tag
            closing_tag = "{% end_cotton_component %}"

            if tag.contents:
                tag_soup = BeautifulSoup(tag.decode_contents(), "html.parser")
                self._transform_components(tag_soup)

                # Create new content with opening tag, tag content, and closing tag
                new_content = opening_tag + str(tag_soup) + closing_tag

            else:
                # Create new content with opening tag and closing tag
                new_content = opening_tag + closing_tag

            # Replace the original tag with the new content
            new_soup = BeautifulSoup(new_content, "html.parser")
            tag.replace_with(new_soup)

        return str(soup)


class UnsortedAttributes(HTMLFormatter):
    def attributes(self, tag):
        for k, v in tag.attrs.items():
            yield k, v


class TemplateCacheHandler:
    def __init__(self):
        self.enabled = getattr(settings, "TEMPLATE_CACHING_ENABLED", True)

    def get_cache_key(self, template_name, mtime):
        template_hash = hashlib.sha256(template_name.encode()).hexdigest()
        return f"cotton_cache_{template_hash}_{mtime}"

    def get_cached_template(self, cache_key):
        if not self.enabled:
            return None
        return cache.get(cache_key)

    def cache_template(self, cache_key, content, timeout=None):
        if self.enabled:
            cache.set(cache_key, content, timeout=timeout)



from django.shortcuts import render

# benchmark tests


def compiled_cotton_test_view(request):
    return render(request, 'compiled_cotton_test.html')


def cotton_test_view(request):
    return render(request, 'cotton_test.cotton.html')


def native_extends_test_view(request):
    return render(request, 'native_extends_test.html')


def native_include_test_view(request):
    return render(request, 'native_include_test.html')

# Django tests


def attribute_merging_test_view(request):
    return render(request, 'attribute_merging_test.cotton.html')


def attribute_passing_test_view(request):
    return render(request, 'attribute_passing_test.cotton.html')